/*
Copyright (C) 2000 Chao-Kuo Lin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "ckDialog.h"
#include "ckTraverser.h"
#include "ckControlManager.h"
#include "ckStream.h"
#include "ckScanner.h"
#include "ckValueWrapper.h"

auto_ptr<ckControlManager> ckControlManager::_instance;


//
// ckControlBuilder
//
ckControlBuilder::ckControlBuilder()
{
    _LSDEBUG1("ckControlBuilder() ctor");
    reset();
}

auto_ptr<ckControl> ckControlBuilder::get() 
{ 
    assert(_log == SUCCESS && _product.get());
    _LSDEBUG1("ckControlBuilder::get()");
    return _product;
}

void ckControlBuilder::reset()
{
    _LSDEBUG1("ckControlBuilder::reset()");
    _log = SUCCESS;
    _current = NULL;
    _top.reset(NULL);
    _product.reset(NULL);
    _projectStarted = false;
}

void ckControlBuilder::startBuilding(string const &type)
{
    assert(_log == SUCCESS);
    _LSDEBUG2("ckControlBuilder::startBuilding() type", type.c_str());
    _projectStarted = true;

    auto_ptr<ckControl> child = ckControl::Create(type);
    ckControl *temp = child.get();
    
    if(temp)
    {
        if(_current)
        {
            assert(_top.get());
            _current->addChild(child);
        }
        else
            _top = child;

        _current = temp;
    }
    else
    {
        _LSDEBUG2("ckControlBuilder::startBuilding() syntax_error", type.c_str());
        ckLogger::instance()->log(SYNTAX_ERROR, type);
        _log = SYNTAX_ERROR;
    }
}

void ckControlBuilder::addAttribute(string const &att, ckValueWrapper &wrapper)
{
    assert(_log == SUCCESS);
    _LSDEBUG2("ckControlBuilder::addAttribute() adding", att.c_str());

    if(_current)
    {
        if(_current->acceptAttribute(att))
        {
            _current->addAttribute(att, wrapper);
        }
        else
        {
            _log = DUPLICATE_DEFINITION;
        }
    }
    else
    {
        _log = ATTRIBUTE_OUTSIDE_NEW;
    }
}

void ckControlBuilder::finishBuilding(string const &type)
{
    assert(_log == SUCCESS);
    _LSDEBUG2("ckControlBuilder::finishBuilding() type", type.c_str());

    if(_current && stricmp(_current->getType().c_str(), type.c_str()) == 0)
    {
        _current = _current->parent();
        if(!_current)
        {
            _LSDEBUG1("ckControlBuilder::finishBuilding() finished building, no active project");
            _product = _top;
            _projectStarted = false;
        }
    }
    else
    {
        _LSDEBUG1("ckControlBuilder::finishBuilding() NEWCONTROL_MISMATCHED");
        ckLogger::instance()->log(NEWCONTROL_MISMATCHED, type);
        _log = NEWCONTROL_MISMATCHED;
    }
}

void ckControlBuilder::terminateBuilding()
{
    if(_current != NULL)
    {
	     if(_current->parent() == NULL)
	     {
		    //assert(stricmp(current->getType().c_str(), DIALOG_TYPE) == 0);
		    _log = NEW_MISMATCHED;
	     }
	     else
	     {
		    _log = NEWCONTROL_MISMATCHED;
	     }

         _LSDEBUG1("ckControlBuilder::terminateBuilding() error outside loop");
    }
}

//
// ckControlManager
//
ckControlManager * ckControlManager::instance()
{
    if(!_instance.get())
    {
        _instance.reset(new ckControlManager);
    }
        
    assert(_instance.get());        
    return _instance.get();
}

void ckControlManager::startBuilding(string const &type)
{
    _LSDEBUG1("ckControlManager::startBuilding() ctor");
    _builder.startBuilding(type);
}

void ckControlManager::addAttribute(string const &att, ckValueWrapper &wrapper)
{
    _LSDEBUG1("we should check for duplicate toplevel ids");
    _LSDEBUG1("ckControlManager::addAttribute() ctor");
    _builder.addAttribute(att, wrapper);
}

void ckControlManager::finishBuilding(string const &type)
{   
    _LSDEBUG1("ckControlManager::finishBuilding() ctor");
    _builder.finishBuilding(type);
    if(checkSyntax() == SUCCESS && !_builder.doingProject())
    {//he just finished a whole block
        _LSDEBUG1("ckControlManager::finishBuilding() finished, pushing in _template");
        _template.push_back(_builder.get());
    }
}



bool ckControlManager::createWindow(string const &id)
{
    _LSDEBUG2("ckControlManager::createWindow() with id: ", id.c_str());
    bool result = false;
    string str, newid;
    ckStrStream stream(id);
    ckDialogTemplateScanner scanner(stream);
    ckValueWrapper wrapper(scanner);
    ckError error = wrapper.getStringWithLength(str, false);
    if(error == SUCCESS)
    {
        error = wrapper.getStringWithLength(newid, false);
        if(error == SUCCESS)
        {
            auto_vector<ckControl>::iterator it = _findChild(str);
            ckCloneTraverser scientist;
            ckControlOneIterator subject(it);

            if(scientist.traverse(subject))
            {
                auto_ptr<ckControl> clone = scientist.getClone();
                assert(clone.get());
        
                ckWindowCreateTraverser traverser;
                //ckDebugTraverser traverser;
                ckControlOneIterator it(clone); // try single iterator

                _LSDEBUG1("ckControlManager::createWindow() about to traverse");
                result = traverser.traverse(it);
                if(result)
                {//only add when it's successful
                    clone.get()->addAttribute(ID_ATTRIB, newid);
                    // we can substibute %[this.blah] into %[newid.blah]
                    ckReplaceEventTraverser eventReplacer(newid);
                    eventReplacer.traverse(it);
                    _active.push_back(clone);
                    _LSDEBUG1("ckControlManager::createWindow() traverse success");
                }
                else
                {
                    _LSDEBUG1("ckControlManager::createWindow() traverse failure");
                }
            }
            else
            {
                ckLogger::instance()->log(FAIL, "Cloning operation failed");
                _LSDEBUG1("ckControlManager::createWindow() failed clone operation");
            }
        }
        else
        {
            ckLogger::instance()->log(SYNTAX_ERROR, "Please specify the new id for the active dialog");
        }
    }

    _LSDEBUGFUNE("ckControlManager::createWindow()");
    return result;
}

//active
// !ckAction invokes this method
ckError ckControlManager::perform(string const &path, string &arguments)
{
    _LSDEBUG2("ckControlManager::perform() with path: ", path.c_str());
    _LSDEBUG2("ckControlManager::perform() with arg: ", arguments.c_str());
    ckStrStream stream(path);
    ckStrStream argstream(arguments);
    ckDialogTemplateScanner scanner(argstream);
    ckValueWrapper wrapper(scanner);
    //ckDebugPathNodeTraverser traverser(stream);
    ckPerformTraverser traverser(stream, wrapper);
    ckControlSingleLevelIterator it(_active);

    _LSDEBUG1("ckControlManager::createWindow() about to traverse");
    
    return traverser.traverse(it) ? SUCCESS : FAIL;
}

//active
ckError ckControlManager::remove(ckControl &control)
{
    _LSDEBUG2("ckControlManager::remove() control ID: ", control.getAttribute(ID_ATTRIB).c_str());
    auto_vector<ckControl>::iterator ait = _active.find(&control);
    if(ait != _active.end())
    {
        _LSDEBUG2("ckControlManager::remove() removed toplevel with id: ", ait->getAttribute(ID_ATTRIB).c_str());
        _active.remove(ait);
    }
    else
    {
        ckRemoveTraverser traverser(control);
        ckControlSingleLevelIterator it(_active);
        if(traverser.traverse(it))
        {//weird return value, but true means not found or not deleted
            _LSDEBUG0;
            assert(false);
        }

        _LSDEBUG1("ckControlManager::remove() removed non toplevel");
    }

    return SUCCESS;
}

// template
ckError ckControlManager::addEventHandler(string const &eventpath, string &handler)
{
    _LSDEBUG2("ckControlManager::addEventHandler() with path: ", eventpath.c_str());
    _LSDEBUG2("ckControlManager::addEventHandler() with handler: ", handler.c_str());
    ckStrStream stream(eventpath);
    ckAddEventTraverser traverser(stream, handler);
    ckControlSingleLevelIterator it(_template);
    return traverser.traverse(it) ? SUCCESS : FAIL;
}

// active
ckError ckControlManager::setVariable(string const &path, string const &value)
{
    _LSDEBUG2("ckControlManager::setAttribute() with path: ", path.c_str());
    _LSDEBUG2("ckControlManager::setAttribute() with value: ", value.c_str());
    ckStrStream stream(path);
    ckSetVariableTraverser traverser(stream, value);
    ckControlSingleLevelIterator it(_active);
    ckControlSingleLevelIterator tit(_template);

    //lovely
    return (traverser.traverse(it) || traverser.traverse(tit)) ? SUCCESS : FAIL;
}

// both
ckError ckControlManager::retreiveVariable(string const &path, string &result)
{
    _LSDEBUG2("ckControlManager::retreiveVariable() with path: ", path.c_str());
    _LSDEBUG2("ckControlManager::retreiveVariable() with result: ", result.c_str());
    ckStrStream stream(path);
    ckRetreiveVariableTraverser traverser(stream, result);
    ckControlSingleLevelIterator it(_active);
    ckControlSingleLevelIterator tit(_template);
    
    // lovely
    return (traverser.traverse(it) || traverser.traverse(tit)) ? SUCCESS : FAIL;
}


auto_vector<ckControl>::iterator ckControlManager::_findChild(string const &id)
{
    auto_vector<ckControl>::iterator it;
    for(it = _template.begin(); it != _template.end(); ++it)
    {
        if(stricmp(it->getAttribute(ID_ATTRIB).c_str(), id.c_str()) == 0)
            break;
    }

    return it;
}

