/*
Copyright (C) 2001 Chao-Kuo Lin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef _CK_TRAVERSER_H_
#define _CK_TRAVERSER_H_

#include "ckDialog.h"
#include "ckControl.h"
#include "ckScanner.h"

class ckPositionableStream;


template<class T>
class ckIterator
{
public:
    ckIterator() {}
    virtual ~ckIterator() {}

    virtual void first() = 0;
    virtual void next() = 0;
    bool virtual isMore() = 0;
    virtual T & get() = 0;
};


//
// ckControlSingleLevelIterator
//
class ckControlSingleLevelIterator : public ckIterator<ckControl>
{
public:
    ckControlSingleLevelIterator(ckControl &control)
    :   _begin(control._children.begin()),
        _current(control._children.begin()),
        _end(control._children.end())
    {
        _LSDEBUG1("ckControlSingleLevelIterator() ctor");
    }

    ckControlSingleLevelIterator(auto_vector<ckControl> &control)
    :   _begin(control.begin()),
        _current(control.begin()),
        _end(control.end())
    {
        _LSDEBUG1("ckControlSingleLevelIterator() ctor");
    }

    ckControlSingleLevelIterator(   auto_vector<ckControl>::iterator begin, 
                                    auto_vector<ckControl>::iterator end)
    :   _begin(begin),
        _end(end),
        _current(begin)
    {
        _LSDEBUG1("ckControlSingleLevelIterator() ctor");
    }

    virtual ~ckControlSingleLevelIterator()
    {
        _LSDEBUG1("ckControlSingleLevelIterator() dtor");
    }

    virtual void first() {   _current = _begin;  }
    virtual void next() { ++_current; }
    virtual bool isMore() { return _current != _end;}

    virtual ckControl & get()
    {
        assert(isMore());
        assert(*_current);
        return *(*_current);
    }

protected:
    auto_vector<ckControl>::iterator _begin, _end, _current;
};

//
// ckControlOneIterator
//
// special case of the single level iterator where you only have one thing
class ckControlOneIterator : public ckControlSingleLevelIterator
{
public:
    ckControlOneIterator(auto_vector<ckControl>::iterator &it)
    :   ckControlSingleLevelIterator(it, ++(auto_vector<ckControl>::iterator(it)))
    {//this way the original it is not changed
        _LSDEBUG1("ckControlSingleLevelIterator() ctor");
    }

    ckControlOneIterator(auto_ptr<ckControl> &control)
    :   ckControlSingleLevelIterator(&control, ++(auto_vector<ckControl>::iterator(&control)))
    {
        _LSDEBUG1("ckControlSingleLevelIterator() ctor");
    }
};

//
// ckControlZeroIterator
//
// special case of the single level iterator where you only have no thing
class ckControlZeroIterator : public ckControlSingleLevelIterator
{
public:
    ckControlZeroIterator()
    :   ckControlSingleLevelIterator(NULL, NULL)
    {
        _LSDEBUG1("ckControlZeroIterator() ctor");
    }
};



//
// ckIdentityCircularIteartor
//
template<class T>
class ckIdentityCircularIteartor : public ckIterator<T>
{
public:
    ckIdentityCircularIteartor(ckIterator<T> &it)
    :   _it(it)
    {
        reset();
    }

    virtual void first() 
    {
        //using this hack
        while(isMore())
            next();
        
        //when there's no more stuff, that
        //means we are back to start
        _wrapper = false;
    }

    virtual void next()
    {
        assert(isMore());
        assert(_it.isMore());
        it.next();

        if(!_it.isMore())
        {
            assert(!_wrapped);
            _wrapped = true;
            _it.first();
        }
    }
    
    bool virtual isMore()
    {
        if(!_t) return false;

        if(_wrapped)
        {
            return _t != &(_it.get());
        }
        else
        {//always more if not wrapped yet
            return true;
        }
    }
    
    virtual T & get() { return _it.get(); }

    // reset the circular list to start from right now
    virtual void reset()
    {
        _wrapped = false;
        if(!_it.isMore())
        {
            _it.first();
            _wrapped = true;
        }

        _t = _it.isMore() ? &(_it.get()) : NULL;
    }

protected:
    ckIterator<T> &_it;
    T *_t;
    bool _wrapped;
};
  



//
// ckTraverser
//
template<class T>
class ckTraverser
{
public:
    ckTraverser() 
    {
        _LSDEBUG1("ckTraverser() ctor");
    }
    
    virtual ~ckTraverser() 
    {
        _LSDEBUG1("ckTraverser() dtor");
    }

    virtual bool traverse(ckIterator<T> &it) = 0;
protected:
    virtual bool _processItem(T &item) = 0;
};

//
// ckControlTraverser
//
class ckControlTraverser : public ckTraverser<ckControl>
{
public:
    virtual bool traverse(ckIterator<ckControl> &it);

protected:
    virtual bool _processItem(ckControl &item) 
    {   return traverse(ckControlSingleLevelIterator(item));    }
};

//
// ckDebugTraverser
//
class ckDebugTraverser : public ckControlTraverser
{
protected:
    virtual bool _processItem(ckControl &item);
};

//
// ckWindowCreateTraverser
//
class ckWindowCreateTraverser : public ckControlTraverser
{
protected:
    virtual bool _processItem(ckControl &item);
};


//
// ckCloneTraverser
//
// a factory traverser, cool eh?
class ckCloneTraverser : public ckControlTraverser
{
public:
    ckCloneTraverser();

    auto_ptr<ckControl> getClone();
    auto_vector<ckControl> getClones();

protected:
    virtual bool _processItem(ckControl &item);

    ckControl *_current;
    auto_vector<ckControl> _specie;
};

//
// ckRemoveTraverser
//
class ckRemoveTraverser : public ckControlTraverser
{
public:
    ckRemoveTraverser(ckControl &control);
    virtual bool _processItem(ckControl &item);

protected:
    ckControl &_control;
};

//
// ckReplaceEventTraverser
//
class ckReplaceEventTraverser : public ckControlTraverser
{
public:
    ckReplaceEventTraverser(string const &rootDialogID);
protected:
    virtual bool _processItem(ckControl &item);
    string _rootDialogID;
};



// createwindow
// cloning

// getcontrol           path
// removechild (probably)


class ckNodePathScanner;
//
// ckControlPathTraverser
//
class ckControlPathTraverser : public ckTraverser<ckControl>
{
public:
    ckControlPathTraverser(ckPositionableStream &stream);
    virtual bool traverse(ckIterator<ckControl> &it);

protected:
    virtual bool _processItem(ckControl &item);
    virtual bool _test(ckControl &item);
    virtual bool _finalProcessing(ckControl &item) = 0;
    
    ckNodePathScanner _scanner;
    ckPositionableStream &_stream;
};


//
// ckDebugPathTraverser
//
class ckDebugPathTraverser : public ckControlPathTraverser 
{
public:
    ckDebugPathTraverser(ckPositionableStream &stream);

protected:
    virtual bool _finalProcessing(ckControl &item);
};

//
// ckGetControlTraverser 
//
class ckGetControlTraverser : public ckControlPathTraverser
{
public:
    ckGetControlTraverser(ckPositionableStream &stream);
    ckControl *getControl() { return _control;}

protected:
    virtual bool _finalProcessing(ckControl &item);

    ckControl *_control;
};

//
// ckControlPathNodeTraverser
//
class ckControlPathNodeTraverser : public ckControlPathTraverser 
{
public:
    ckControlPathNodeTraverser(ckPositionableStream &stream);

protected:
    virtual bool _finalProcessing(ckControl &item);
    virtual bool _match(ckControl &item, string const &str) = 0;
};


//
// ckDebugPathNodeTraverser
//
class ckDebugPathNodeTraverser : public ckControlPathNodeTraverser
{
public:
    ckDebugPathNodeTraverser(ckPositionableStream &stream);

protected:
    virtual bool _match(ckControl &item, string const &str);
};

//
// ckAddEventTraverser
//
class ckAddEventTraverser : public ckControlPathNodeTraverser
{
public:
    ckAddEventTraverser(ckPositionableStream &stream, string const &action);

protected:
    virtual bool _match(ckControl &item, string const &str);

    const string &_action;
};

//
// ckNotifyTraverser
//
class ckNotifyTraverser : public ckControlPathNodeTraverser
{
public:
    ckNotifyTraverser(ckPositionableStream &stream);

protected:
    virtual bool _match(ckControl &item, string const &str);
};

//
// ckSetAttributeTraverser 
//
class ckSetVariableTraverser : public ckControlPathNodeTraverser
{
public:
    ckSetVariableTraverser(ckPositionableStream &stream, string const &value);

protected:
    virtual bool _match(ckControl &item, string const &str);
    string _value;
};

//
// ckPerformTraverser 
//
class ckPerformTraverser : public ckControlPathNodeTraverser
{
public:
    ckPerformTraverser(ckPositionableStream &stream, ckValueWrapper &wrapper);

protected:
    virtual bool _match(ckControl &item, string const &str);

    ckValueWrapper &_wrapper;
};

//
// ckRetreiveVariableTraverser 
//
class ckRetreiveVariableTraverser : public ckControlPathNodeTraverser
{
public:
    ckRetreiveVariableTraverser(ckPositionableStream &stream, string &result);


protected:
    virtual bool _match(ckControl &item, string const &str);

    string &_result;
};

#endif


